import time
from dataclasses import dataclass, field

from .event import Event
from .key import PrivateKey

BECH32_CHARS = '023456789acdefghjklmnpqrstuvwxyz'


def zero_bits(b: int) -> int:
    n = 0

    if b == 0:
        return 8

    while b >> 1:
        b = b >> 1
        n += 1

    return 7 - n


def count_leading_zero_bits(hex_str: str) -> int:
    total = 0
    for i in range(0, len(hex_str) - 2, 2):
        bits = zero_bits(int(hex_str[i : i + 2], 16))
        total += bits

        if bits != 8:
            break

    return total


def _guess_event(event: Event) -> Event:
    num_leading_zero_bits = count_leading_zero_bits(event.id)
    return num_leading_zero_bits, event


def _guess_key():
    sk = PrivateKey()
    num_leading_zero_bits = count_leading_zero_bits(sk.public_key.hex())
    return num_leading_zero_bits, sk


def _guess_vanity_key():
    sk = PrivateKey()
    vk = sk.public_key.bech32()
    return vk, sk


@dataclass
class Pow:
    def __post_init__(self):
        self.count = 0
        self.results = []
        self.duration = 0
        self.mode = "key"
        self.num_leading_zero_bits = 0
        self.n_pattern = 0
        self.n_options = 2
        self.operation = None

    def get_hashrate(self):
        if self.duration > 0 and self.count > 0:
            return self.count / self.duration
        else:
            return 0

    def _stop_mining(self, count, max_count, duration, max_duration):
        if max_count > 0 and count > max_count:
            return True
        elif max_duration > 0 and duration > max_duration:
            return True
        return False

    def estimate_hashrate(self, n_guesses: int = 1e4, **operation_kwargs):
        if self.operation is None:
            return 0
        n_guesses = int(n_guesses)

        def _time_operation():
            start = time.perf_counter()
            self.operation(**operation_kwargs)
            end = time.perf_counter()
            return end - start

        t = sum([_time_operation() for _ in range(n_guesses)]) / n_guesses
        hashrate = 1 / t
        return hashrate

    def get_expected_guesses(self):
        p = 1 / self.n_options
        return 1 / (p**self.n_pattern)

    def stored_results(self):
        return f"{len(self.results)}. {self.results[-1]}"

    def print_results(self):
        if self.mode in ["key", "event"]:
            print(
                f"Found event with {self.num_leading_zero_bits} "
                f"zeros after {self.count} tries"
            )
        print(f"avg. hash rate {self.get_hashrate()} guesses per s")
        print(f"Total duration {self.duration} s")


@dataclass
class PowEvent(Pow):
    difficulty: int = 8
    event: Event = field(default_factory=Event)

    def __post_init__(self):
        self.mode = "event"
        self.operation = _guess_event
        self.n_options = 2
        self.n_pattern = self.difficulty
        self.reset()

    def reset(self):
        event = Event(
            content=self.event.content,
            public_key=self.event.public_key,
            kind=self.event.kind,
        )
        all_tags = [["nonce", "1", str(self.difficulty)]]
        all_tags.extend(self.event.tags)
        event.tags = all_tags
        self.num_leading_zero_bits, self.event = self.operation(event)
        self.count = 1
        self.duration = 0
        self.results = []

    def set_difficulty(self, difficulty):
        self.difficulty = difficulty
        self.n_pattern = self.difficulty
        self.event.tags[0][2] = str(self.difficulty)

    def increase_difficulty(self):
        self.set_difficulty(self.num_leading_zero_bits + 1)

    def mine(self, max_count: int = 0, max_duration: int = 0) -> Event:
        start = time.perf_counter()
        count = 0
        duration = 0
        num_leading_zero_bits = self.num_leading_zero_bits
        event = self.event
        while num_leading_zero_bits < self.difficulty and (
            not self._stop_mining(count, max_count, duration, max_duration)
        ):
            self.count += 1
            event.tags[0][1] = str(self.count)
            num_leading_zero_bits, event = self.operation(event)
            if num_leading_zero_bits > self.num_leading_zero_bits:
                self.event = event
                self.num_leading_zero_bits = num_leading_zero_bits
            count += 1
            duration = time.perf_counter() - start
        end = time.perf_counter()
        self.duration += end - start
        self.results.append((self.num_leading_zero_bits, self.event))
        return self.event

    def get_expected_time(self, hashrate=None) -> float:
        if hashrate is None:
            if self.count > 10000 and self.duration > 0:
                hashrate = self.get_hashrate()
            else:
                hashrate = self.estimate_hashrate(event=self.event)
        self.n_pattern = self.difficulty
        self.n_options = 2
        return self.get_expected_guesses() / hashrate


@dataclass
class PowKey(Pow):
    difficulty: int = 8

    def __post_init__(self):
        self.mode = "key"
        self.operation = _guess_key
        self.n_options = 2
        self.n_pattern = self.difficulty
        self.reset()

    def reset(self):
        self.num_leading_zero_bits, self.sk = self.operation()
        self.count = 0
        self.duration = 0
        self.results = []

    def set_difficulty(self, difficulty):
        self.difficulty = difficulty
        self.n_pattern = difficulty

    def increase_difficulty(self):
        self.set_difficulty(self.num_leading_zero_bits + 1)

    def mine(self, max_count: int = 0, max_duration: int = 0) -> PrivateKey:
        start = time.perf_counter()
        count = 0
        duration = 0
        num_leading_zero_bits = self.num_leading_zero_bits
        while num_leading_zero_bits < self.difficulty and (
            not self._stop_mining(count, max_count, duration, max_duration)
        ):
            num_leading_zero_bits, sk = self.operation()
            if num_leading_zero_bits > self.num_leading_zero_bits:
                self.sk = sk
                self.num_leading_zero_bits = num_leading_zero_bits
            count += 1
            duration = time.perf_counter() - start
        self.count += count
        end = time.perf_counter()
        self.duration += end - start
        self.results.append((self.num_leading_zero_bits, self.sk))
        return self.sk

    def get_expected_time(self, hashrate=None) -> float:
        if hashrate is None:
            if self.count > 10000 and self.duration > 0:
                hashrate = self.get_hashrate()
            else:
                hashrate = self.estimate_hashrate()
        self.n_pattern = self.difficulty
        self.n_options = 2
        return self.get_expected_guesses() / hashrate


@dataclass
class PowVanityKey(Pow):
    prefix: str = None
    suffix: str = None

    def __post_init__(self):
        self.n_pattern = 0
        self.operation = _guess_vanity_key
        self.n_options = len(BECH32_CHARS)
        self.mode = "vanity_key"
        if self.prefix is None and self.suffix is None:
            raise ValueError("Expected at least one of 'prefix' or 'suffix' arguments")

        for pattern in [self.prefix, self.suffix]:
            if pattern is not None:
                self.n_pattern += len(pattern)
                missing_chars = [c for c in pattern if c not in BECH32_CHARS]
                if len(missing_chars) > 0:
                    raise ValueError(
                        f"{missing_chars} not in valid "
                        f"list of bech32 chars: ({BECH32_CHARS})"
                    )
        self.reset()

    def reset(self):
        self.count = 0
        self.duration = 0
        self.results = []

    def _check_vanity(self):
        if (
            self.prefix is not None
            and not self.vk[5 : 5 + len(self.prefix)] == self.prefix
        ):
            return False
        if self.suffix is not None and not self.vk[-len(self.suffix) :] == self.suffix:
            return False
        return True

    def mine(self, max_count: int = 0, max_duration: int = 0) -> PrivateKey:
        start = time.perf_counter()
        count = 0
        duration = 0
        self.vk, self.sk = self.operation()
        while not self._check_vanity() and (
            not self._stop_mining(count, max_count, duration, max_duration)
        ):
            self.vk, self.sk = self.operation()
            count += 1
            duration = time.perf_counter() - start
        self.count += count
        end = time.perf_counter()
        self.duration += end - start
        if self._check_vanity():
            self.results.append((self.vk, self.sk))
        return self.sk

    def get_expected_time(self, hashrate=None):
        if hashrate is None:
            if self.count > 10000 and self.duration > 0:
                hashrate = self.get_hashrate()
            else:
                hashrate = self.estimate_hashrate()
        return self.get_expected_guesses() / hashrate



import json
import time
import uuid

import click

from nostr.commands.config import Config
from nostr.event import EncryptedDirectMessage, Event, EventKind
from nostr.filter import Filter, Filters
from nostr.key import PrivateKey, PublicKey
from nostr.message_type import ClientMessageType
from nostr.relay_manager import RelayManager
from nostr.utils import dict2obj


@click.group()
@click.option("-c", "--config", required=False, type=str, help="Config file")
@click.pass_context
def cli(ctx, config: str = None):
    """Command related to message(s)."""
    ctx.ensure_object(dict)

    # Set default relays
    ctx.obj['relays'] = ["wss://nostr-pub.wellorder.net", "wss://relay.damus.io"]

    config = Config.load(config)
    if config:
        try:
            relays = dict2obj(config).nostr[0].relays
            if relays:
                ctx.obj['relays'] = relays
        except AttributeError:
            pass

        try:
            influencers = dict2obj(config).nostr[0].listen
            if influencers:
                ctx.obj['influencers'] = influencers
        except AttributeError:
            pass

        try:
            ctx.obj['self'] = dict2obj(config).nostr[0].self[0]
        except AttributeError:
            pass

        try:
            receivers = dict2obj(config).nostr[0].receiver
            if receivers:
                ctx.obj['receivers'] = receivers
        except AttributeError:
            pass


@cli.command()
@click.option("-i", "--identifier", required=False, type=str)
@click.option("-p", "--pub-key", "npub", required=False, type=str)
@click.option("-l", "--limit", "limit", type=int, default=10)
@click.option("-s", "--sleep", "sleep", type=int, default=2)
@click.pass_context
def receive(ctx: dict, identifier: str, npub: str, limit: int = 10, sleep: int = 2):
    """Receives messages from npub address."""
    npubs = [npub] if npub else []
    if identifier:
        for influencer in ctx.obj.get('influencers', []):
            if hasattr(influencer, identifier):
                try:
                    npubs.append(getattr(influencer, identifier).npub)
                except AttributeError:
                    pass

    if not npubs:
        print(
            "Please input `npub`; and/or",
            "specify listen `identifier` by creating the config file",
        )
        return 1

    authors = [PublicKey.from_npub(npub).hex() for npub in npubs]
    filters = Filters(
        [Filter(authors=authors, kinds=[EventKind.TEXT_NOTE], limit=limit)]
    )
    subscription_id = uuid.uuid1().hex
    request = [ClientMessageType.REQUEST, subscription_id]
    request.extend(filters.to_json_array())

    relay_manager = RelayManager()
    for relay in ctx.obj['relays']:
        relay_manager.add_relay(relay)
    relay_manager.add_subscription(subscription_id, filters)

    def get_events(relay_manager):
        events = []
        while relay_manager.message_pool.has_events():
            event_msg = relay_manager.message_pool.get_event()
            events.append(event_msg.event.content)
        return events

    def get_notices(relay_manager):
        notices = []
        while relay_manager.message_pool.has_notices():
            notice_msg = relay_manager.message_pool.get_notice()
            notices.append(notice_msg.content)
        return notices

    with relay_manager:
        time.sleep(sleep)  # allow the connections to open
        message = json.dumps(request)
        relay_manager.publish_message(message)
        time.sleep(sleep)  # allow the messages to send

        events = get_events(relay_manager=relay_manager)
        notices = get_notices(relay_manager=relay_manager)
        click.echo(
            json.dumps(
                {"Public key(s)": npubs, "Events": events, "Notices": notices}, indent=2
            )
        )
    return 0


@cli.command()
@click.option("-s", "--sec-key", "nsec", required=False, type=str)
@click.option("-m", "--message", "message", type=str)
@click.option("--sleep", "sleep", type=int, default=2)
@click.pass_context
def publish(ctx: dict, nsec: str, message: str, sleep: int = 2):
    """Sends a message."""
    if not nsec and ctx.obj.get('self'):
        try:
            nsec = ctx.obj.get('self').nsec
        except AttributeError:
            pass

    if not nsec:
        print(
            "Please input `nsec`; and/or",
            "set `nsec`, under `self` block in config file",
        )
        return 1
    private_key = PrivateKey.from_nsec(nsec)
    event = Event(content=message, public_key=private_key.public_key.hex())
    event.sign(private_key.hex())

    msg = json.dumps([ClientMessageType.EVENT, event.to_dict()])

    relay_manager = RelayManager()
    for relay in ctx.obj['relays']:
        relay_manager.add_relay(relay)
    with relay_manager:
        time.sleep(sleep)  # allow the connections to open
        relay_manager.publish_message(msg)

    click.echo(json.dumps({"Message": message}, indent=2))
    return 0


@cli.command()
@click.option("-s", "--sec-key", "nsec", required=False, type=str)
@click.option("-m", "--message", "message", type=str)
@click.option("-i", "--identifier", required=False, type=str)
@click.option("-p", "--pub-key", "receiver_npub", required=False, type=str)
@click.option("--sleep", "sleep", type=int, default=2)
@click.pass_context
def send(
    ctx: dict,
    nsec: str,
    message: str,
    identifier: str,
    receiver_npub: str,
    sleep: int = 2,
):
    """Sends a encryped direct message."""
    if not nsec and ctx.obj.get('self'):
        try:
            nsec = ctx.obj.get('self').nsec
        except AttributeError:
            pass

    if not nsec:
        print(
            "Please input `nsec`; and/or",
            "set `nsec`, under `self` block in config file",
        )
        return 1

    receiver_npub = receiver_npub or None
    if not receiver_npub and identifier:
        for receiver in ctx.obj.get('receivers', []):
            if hasattr(receiver, identifier):
                try:
                    receiver_npub = getattr(receiver, identifier).npub
                except AttributeError:
                    pass

    if not receiver_npub:
        print(
            "Please input `receiver_npub`; and/or",
            "specify `identifier`, which should be found",
            "under `receiver` block in config file",
        )
        return 1

    private_key = PrivateKey.from_nsec(nsec)
    recipient_pubkey = PublicKey.from_npub(receiver_npub)
    direct_message = EncryptedDirectMessage(
        recipient_pubkey=recipient_pubkey.hex(), cleartext_content=message
    )
    direct_message.encrypt_dm(
        private_key_hex=private_key.hex(),
        cleartext_content=message,
        recipient_pubkey=recipient_pubkey.hex(),
    )
    direct_message.sign(private_key.hex())

    relay_manager = RelayManager()
    for relay in ctx.obj['relays']:
        relay_manager.add_relay(relay)

    with relay_manager:
        time.sleep(sleep)  # allow the connections to open
        relay_manager.publish_event(direct_message)

    click.echo(json.dumps({"Message": message}, indent=2))
    return 0



class ClientMessageType:
    EVENT = "EVENT"
    REQUEST = "REQ"
    CLOSE = "CLOSE"


class RelayMessageType:
    EVENT = "EVENT"
    NOTICE = "NOTICE"
    OK = "OK"
    END_OF_STORED_EVENTS = "EOSE"

    @staticmethod
    def is_valid(type: str) -> bool:
        if (
            type == RelayMessageType.EVENT
            or type == RelayMessageType.NOTICE
            or type == RelayMessageType.END_OF_STORED_EVENTS
            or type == RelayMessageType.OK
        ):
            return True
        return False



from .filter import Filters


class Subscription:
    def __init__(self, id: str, filters: Filters = None, batch: int = 0) -> None:
        self.id = id
        self.filters = filters
        self.batch = batch
        self.paused = False

        if not isinstance(self.id, str):
            raise TypeError("Argument 'id' must be of type str")

    def to_json_object(self):
        return {
            "id": self.id,
            "filters": self.filters.to_json_array(),
            "batch": self.batch,
            "paused": self.paused,
        }



import json
import ssl
import threading
import time
from dataclasses import dataclass
from threading import Lock

from .event import Event
from .filter import Filters
from .message_pool import MessagePool
from .relay import Relay, RelayPolicy, RelayProxyConnectionConfig
from .request import Request


class RelayException(Exception):
    pass


@dataclass
class RelayManager:
    error_threshold: int = 0

    def __post_init__(self):
        self.relays: dict[str, Relay] = {}
        self.message_pool: MessagePool = MessagePool()
        self.lock: Lock = Lock()

    def add_relay(
        self,
        url: str,
        policy: RelayPolicy = RelayPolicy(),
        ssl_options: dict = None,
        proxy_config: RelayProxyConnectionConfig = None,
    ):
        relay = Relay(url, self.message_pool, policy, ssl_options, proxy_config)
        if self.error_threshold:
            relay.error_threshold = self.error_threshold

        with self.lock:
            self.relays[url] = relay

    def remove_relay(self, url: str):
        with self.lock:
            if url in self.relays:
                relay = self.relays.pop(url)
                relay.close()

    def remove_closed_relays(self):
        for url, connected in self.connection_statuses.items():
            if not connected:
                # warnings.warn(f'{url} is not connected... removing relay.')
                self.remove_relay(url=url)

    def add_subscription(self, id: str, filters: Filters):
        with self.lock:
            for relay in self.relays.values():
                if relay.policy.should_read:
                    relay.add_subscription(id, filters)

    def close_subscription(self):
        for relay in self.relays.values():
            relay.close_subscription(id)

    def __enter__(self):
        # NOTE: This disables ssl certificate verification
        self.open_connections({"cert_reqs": ssl.CERT_NONE})

    def open_connections(self, ssl_options: dict = None):
        for relay in self.relays.values():
            if not relay.is_connected:
                threading.Thread(
                    target=relay.connect,
                    name=f"{relay.url}-thread",
                ).start()
        time.sleep(2)
        self.remove_closed_relays()
        assert all(self.connection_statuses.values())

    def __exit__(self, type, value, traceback):
        self.close_connections()

    def close_connections(self):
        for relay in self.relays.values():
            relay.close()

        assert not any(self.connection_statuses.values())

    def get_connection_status(self):
        out = []
        for relay in self.relays.values():
            out.append([relay.url, relay.active])
        return out

    def add_subscription_on_relay(self, url: str, id: str, filters: Filters):
        with self.lock:
            if url in self.relays:
                relay = self.relays[url]
                if not relay.policy.should_read:
                    raise RelayException(
                        f"Could not send request: {url} \
                        is not configured to read from"
                    )
                relay.add_subscription(id, filters)
                request = Request(id, filters)
                relay.publish(request.to_message())
            else:
                raise RelayException(f"Invalid relay url: no connection to {url}")

    def add_subscription_on_all_relays(self, id: str, filters: Filters):
        with self.lock:
            for relay in self.relays.values():
                if relay.policy.should_read:
                    relay.add_subscription(id, filters)
                    request = Request(id, filters)
                    relay.publish(request.to_message())

    def close_subscription_on_relay(self, url: str, id: str):
        with self.lock:
            if url in self.relays:
                relay = self.relays[url]
                relay.close_subscription(id)
                relay.publish(json.dumps(["CLOSE", id]))
            else:
                raise RelayException(f"Invalid relay url: no connection to {url}")

    def close_subscription_on_all_relays(self, id: str):
        with self.lock:
            for relay in self.relays.values():
                relay.close_subscription(id)
                relay.publish(json.dumps(["CLOSE", id]))

    def close_all_relay_connections(self):
        with self.lock:
            for url in self.relays:
                relay = self.relays[url]
                relay.close()

    @property
    def connection_statuses(self) -> dict:
        """gets the url and connection statuses of relays
        Returns:
            dict: bool of connection statuses
        """
        statuses = [relay.is_connected for relay in self.relays.values()]
        return dict(zip(self.relays.keys(), statuses))

    def publish_message(self, message: str):
        with self.lock:
            for relay in self.relays.values():
                if relay.policy.should_write:
                    relay.publish(message)

    def publish_event(self, event: Event):
        """Verifies that the Event is publishable before submitting it to relays."""
        if event.signature is None:
            raise RelayException(f"Could not publish {event.id}: must be signed")

        if not event.verify():
            raise RelayException(
                f"Could not publish {event.id}: \
                failed to verify signature {event.signature}"
            )

        self.publish_message(event.to_message())



import json
import unittest
from pathlib import Path
from unittest import mock
from unittest.mock import ANY, MagicMock, patch

import hcl2
from click.testing import CliRunner

from nostr.commands.message import cli
from nostr.event import Event
from nostr.message_pool import EventMessage


class TestCLIMessageWithConfig(unittest.TestCase):
    def setUp(self):
        self.config_file = Path.cwd().joinpath('test', 'fixtures', 'config.hcl')
        with open(self.config_file) as file:
            self.config = hcl2.load(file)

    @patch('nostr.commands.message.RelayManager', autospec=True)
    def test_publish(self, mock_relay_manager):
        # GIVEN
        message = "Hello nostr world!!"
        runner = CliRunner()

        mock_manager = MagicMock()
        mock_relay_manager.return_value = mock_manager
        mock_manager.__enter__.return_value = mock_manager
        mock_manager.add_relay.return_value = None

        # WHEN
        result = runner.invoke(
            cli,
            ['-c', self.config_file, 'publish', '-m', message, '--sleep', 0],
            catch_exceptions=False,
        )

        # THEN
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(json.loads(result.output), {"Message": message})
        mock_manager.add_relay.assert_has_calls([mock.call(ANY), mock.call(ANY)])
        mock_manager.publish_message.assert_called()

    @patch('nostr.commands.message.RelayManager', autospec=True)
    def test_send(self, mock_relay_manager):
        # GIVEN
        message = "Hello nostr world!!"
        runner = CliRunner()

        mock_manager = MagicMock()
        mock_relay_manager.return_value = mock_manager
        mock_manager.__enter__.return_value = mock_manager
        mock_manager.add_relay.return_value = None

        # WHEN
        result = runner.invoke(
            cli,
            ['-c', self.config_file, 'send', '-m', message, '-i', 'Ray', '--sleep', 0],
        )

        # THEN
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(json.loads(result.output), {"Message": message})
        mock_manager.add_relay.assert_has_calls([mock.call(ANY), mock.call(ANY)])
        mock_manager.publish_event.assert_called()

    @patch('nostr.commands.message.RelayManager', autospec=True)
    def test_receive(self, mock_relay_manager):
        # GIVEN
        runner = CliRunner()

        mock_manager = MagicMock()
        mock_relay_manager.return_value = mock_manager
        mock_manager.__enter__.return_value = mock_manager
        mock_manager.add_relay.return_value = None

        mock_manager.message_pool.has_events.side_effect = [True, False]
        mock_manager.message_pool.get_event.return_value = EventMessage(
            Event(content="any"), subscription_id=123, url="wss://any.relay"
        )

        mock_manager.message_pool.has_notices.return_value = False

        # WHEN
        result = runner.invoke(
            cli,
            ['-c', self.config_file, 'receive', '-i', 'jon', '-s', 0],
        )

        # THEN
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(
            json.loads(result.output),
            {
                "Public key(s)": [self.config['nostr'][0]['listen'][0]['jon']['npub']],
                "Events": ANY,
                "Notices": ANY,
            },
        )
        mock_manager.add_relay.assert_has_calls([mock.call(ANY), mock.call(ANY)])
        mock_manager.publish_message.assert_called()
        mock_manager.message_pool.has_events.assert_called()
        mock_manager.message_pool.get_event.assert_called()
        mock_manager.message_pool.has_notices.assert_called()
        mock_manager.message_pool.get_notice.assert_not_called()



import binascii
import secrets
from base64 import b64decode, b64encode
from hashlib import sha256
from typing import Optional

import coincurve as secp256k1
from coincurve._libsecp256k1 import ffi, lib
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from . import bech32
from .delegation import Delegation

# from .event import EncryptedDirectMessage, Event, EventKind

HAS_ECDH = hasattr(lib, "secp256k1_ecdh")


class PublicKey:
    def __init__(self, raw_bytes: bytes) -> None:
        """
        :param raw_bytes: The formatted public key.
        :type data: bytes
        """
        if isinstance(raw_bytes, PrivateKey):
            self.raw_bytes = raw_bytes.public_key.raw_bytes
        elif isinstance(raw_bytes, secp256k1.keys.PublicKey):
            self.raw_bytes = raw_bytes.format(compressed=True)[2:]
        elif isinstance(raw_bytes, secp256k1.keys.PublicKeyXOnly):
            self.raw_bytes = raw_bytes.format()
        elif isinstance(raw_bytes, str):
            self.raw_bytes = binascii.unhexlify(raw_bytes)
        else:
            self.raw_bytes = raw_bytes

    def bech32(self) -> str:
        converted_bits = bech32.convertbits(self.raw_bytes, 8, 5)
        return bech32.bech32_encode("npub", converted_bits, bech32.Encoding.BECH32)

    def hex(self) -> str:
        return self.raw_bytes.hex()

    def verify_signed_message_hash(self, hash: str, sig: str) -> bool:
        pk = secp256k1.PublicKey(b"\x02" + self.raw_bytes, True)
        return pk.schnorr_verify(bytes.fromhex(hash), bytes.fromhex(sig), None, True)

    def verify(self, sig: bytes, message: bytes) -> bool:
        pk = secp256k1.PublicKeyXOnly(self.raw_bytes)
        return pk.verify(sig, message)

    @classmethod
    def from_hex(cls, hex: str) -> "PublicKey":
        return cls(bytes.fromhex(hex))

    @classmethod
    def from_npub(cls, npub: str):
        """Load a PublicKey from its bech32/npub form."""
        hrp, data, spec = bech32.bech32_decode(npub)
        raw_public_key = bech32.convertbits(data, 5, 8)[:-1]
        return cls(bytes(raw_public_key))


class PrivateKey:
    def __init__(self, raw_secret: Optional[bytes] = None) -> None:
        if raw_secret:
            self.raw_secret = raw_secret
        else:
            self.raw_secret = secrets.token_bytes(32)

        sk = secp256k1.PrivateKey(self.raw_secret)
        self.public_key = PublicKey(sk.public_key_xonly)

    @classmethod
    def from_nsec(cls, nsec: str):
        """Load a PrivateKey from its bech32/nsec form."""
        hrp, data, spec = bech32.bech32_decode(nsec)
        raw_secret = bech32.convertbits(data, 5, 8)[:-1]
        return cls(bytes(raw_secret))

    @classmethod
    def from_hex(cls, hex: str):
        """Load a PrivateKey from its hex form."""
        return cls(binascii.unhexlify(hex))

    def bech32(self) -> str:
        converted_bits = bech32.convertbits(self.raw_secret, 8, 5)
        return bech32.bech32_encode("nsec", converted_bits, bech32.Encoding.BECH32)

    def hex(self) -> str:
        return self.raw_secret.hex()

    def ecdh(self, public_key_hex):
        assert public_key_hex, "No public key defined"
        if not HAS_ECDH:
            raise Exception("secp256k1_ecdh not enabled")
        sk = secp256k1.PrivateKey(self.raw_secret)
        result = ffi.new("char [32]")
        pk = secp256k1.PublicKey(bytes.fromhex("02" + public_key_hex))
        res = lib.secp256k1_ecdh(
            sk.context.ctx, result, pk.public_key, self.raw_secret, copy_x, ffi.NULL
        )
        if not res:
            raise Exception(f"invalid scalar ({res})")

        return bytes(ffi.buffer(result, 32))

    def tweak_add(self, scalar: bytes) -> bytes:
        sk = secp256k1.PrivateKey(self.raw_secret)
        return sk.tweak_add(scalar)

    def compute_shared_secret(self, public_key_hex: str) -> bytes:
        return self.ecdh(public_key_hex)
        # pk = secp256k1.PublicKey(bytes.fromhex("02" + public_key_hex), True)
        # return pk.ecdh(self.raw_secret, hashfn=copy_x)

    def encrypt_message(self, message: str, public_key_hex: str) -> str:
        padder = padding.PKCS7(128).padder()
        padded_data = padder.update(message.encode()) + padder.finalize()

        iv = secrets.token_bytes(16)
        cipher = Cipher(
            algorithms.AES(self.compute_shared_secret(public_key_hex)), modes.CBC(iv)
        )

        encryptor = cipher.encryptor()
        encrypted_message = encryptor.update(padded_data) + encryptor.finalize()

        return f"{b64encode(encrypted_message).decode()}?iv={b64encode(iv).decode()}"

    # def encrypt_dm(self, dm: EncryptedDirectMessage) -> None:
    #     dm.content = self.encrypt_message(
    #         message=dm.cleartext_content, public_key_hex=dm.recipient_pubkey
    #     )

    def decrypt_message(self, encoded_message: str, public_key_hex: str) -> str:
        encoded_data = encoded_message.split("?iv=")
        encoded_content, encoded_iv = encoded_data[0], encoded_data[1]

        iv = b64decode(encoded_iv)
        cipher = Cipher(
            algorithms.AES(self.compute_shared_secret(public_key_hex)), modes.CBC(iv)
        )
        encrypted_content = b64decode(encoded_content)

        decryptor = cipher.decryptor()
        decrypted_message = decryptor.update(encrypted_content) + decryptor.finalize()

        unpadder = padding.PKCS7(128).unpadder()
        unpadded_data = unpadder.update(decrypted_message) + unpadder.finalize()

        return unpadded_data.decode()

    def sign_message_hash(self, hash: bytes) -> str:
        sk = secp256k1.PrivateKey(self.raw_secret)
        sig = sk.schnorr_sign(hash, None, raw=True)
        return sig.hex()

    def sign(self, message: bytes, aux_randomness: bytes = b"") -> str:
        sk = secp256k1.PrivateKey(self.raw_secret)
        return sk.sign_schnorr(message, aux_randomness)

    def sign_delegation(self, delegation: Delegation) -> None:
        delegation.signature = self.sign_message_hash(
            sha256(delegation.delegation_token.encode()).digest()
        )

    def __eq__(self, other):
        return self.raw_secret == other.raw_secret


def mine_vanity_key(
    prefix: Optional[str] = None, suffix: Optional[str] = None
) -> PrivateKey:
    if prefix is None and suffix is None:
        raise ValueError("Expected at least one of 'prefix' or 'suffix' arguments")

    while True:
        sk = PrivateKey()
        if (
            prefix is not None
            and not sk.public_key.bech32()[5 : 5 + len(prefix)] == prefix
        ):
            continue
        if suffix is not None and not sk.public_key.bech32()[-len(suffix) :] == suffix:
            continue
        break

    return sk


@ffi.callback(
    "int (unsigned char *, const unsigned char *, const unsigned char *, void *)"
)
def copy_x(output, x32, y32, data):
    ffi.memmove(output, x32, 32)
    return 1



import json
import time
from dataclasses import dataclass, field
from enum import IntEnum
from hashlib import sha256
from typing import List, Optional

from .key import PrivateKey, PublicKey
from .message_type import ClientMessageType


class EventKind(IntEnum):
    SET_METADATA = 0
    TEXT_NOTE = 1
    RECOMMEND_RELAY = 2
    CONTACTS = 3
    ENCRYPTED_DIRECT_MESSAGE = 4
    DELETE = 5
    BOOST = 6
    REACTION = 7
    CHANNEL_CREATE = 40
    CHANNEL_META = 41
    CHANNEL_MESSAGE = 42
    CHANNEL_HIDE = 43
    CHANNEL_MUTE = 44
    RELAY_LIST_METADATA = 10002


@dataclass
class Event:
    """Event class.

    :param content: content string
    :param public_key: public key in hex form
    :param created_at: event creation date
    :param kind: event kind
    :param tags: list of list of strings
    :param signature: signature, will be created after signing with a private key
    """

    content: Optional[str] = None
    public_key: Optional[str] = None
    created_at: Optional[int] = None
    kind: Optional[int] = EventKind.TEXT_NOTE
    tags: List[List[str]] = field(default_factory=list)
    signature: Optional[str] = None

    def __post_init__(self) -> None:
        if self.content and not isinstance(self.content, str):
            raise TypeError("Argument 'content' must be of type str")
        # elif (
        #     self.content
        #     and "?iv" not in self.content
        #     and self.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE
        # ):
        #     raise Exception(
        #         "Encrypted DMs cannot use the `content` field; use encrypt_dm()"
        #         "for storing an encrypted content."
        #     )

        if self.created_at is None:
            self.created_at = int(time.time())

    @staticmethod
    def serialize(
        public_key: str, created_at: int, kind: int, tags: List[List[str]], content: str
    ) -> bytes:
        data = [0, public_key, created_at, kind, tags, content]
        data_str = json.dumps(data, separators=(",", ":"), ensure_ascii=False)
        return data_str.encode()

    @staticmethod
    def compute_id(
        public_key: str, created_at: int, kind: int, tags: List[List[str]], content: str
    ):
        return sha256(
            Event.serialize(public_key, created_at, kind, tags, content)
        ).hexdigest()

    @property
    def id(self) -> str:
        # Always recompute the id to reflect the up-to-date state of the Event
        return Event.compute_id(
            self.public_key, self.created_at, self.kind, self.tags, self.content
        )

    def encrypt_dm(
        self, private_key_hex: str, cleartext_content: str, recipient_pubkey: str
    ) -> None:
        if self.kind != EventKind.ENCRYPTED_DIRECT_MESSAGE:
            raise Exception("Wrong event kind, needs to be ENCRYPTED_DIRECT_MESSAGE")
        if not self.has_pubkey_ref(recipient_pubkey):
            # Must specify the DM recipient's pubkey in a 'p' tag
            self.add_pubkey_ref(recipient_pubkey)
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        encrypted_message = sk.encrypt_message(
            message=cleartext_content, public_key_hex=recipient_pubkey
        )
        self.content = encrypted_message

    def sign(self, private_key_hex: str) -> None:
        """signs the event with the private key and stored the signature in self.sig.
        The pubkey from the event is replaced and the note id recomputed.
        :param private_key_hex: private key as hex string
        """
        if self.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE and self.content is None:
            raise Exception("Message is not yet encrypted!")
        sk = PrivateKey(bytes.fromhex(private_key_hex))
        self.public_key = sk.public_key.hex()
        sig = sk.sign(bytes.fromhex(self.id))
        self.signature = sig.hex()

    def add_pubkey_ref(self, pubkey: str):
        """Adds a reference to a pubkey as a 'p' tag."""
        self.tags.append(["p", pubkey])

    def has_pubkey_ref(self, pubkey: str) -> bool:
        """Check if a `p` tag to the given pubkey exists."""
        for tag_type, tag in self.tags:
            if tag_type == "p" and tag == pubkey:
                return True
        return False

    def add_event_ref(self, event_id: str):
        """Adds a reference to an event_id as an 'e' tag."""
        self.tags.append(["e", event_id])

    def has_event_ref(self, event_id: str):
        """Check if a `e` tag to the given event_id exists."""
        for tag_type, tag in self.tags:
            if tag_type == "e" and tag == event_id:
                return True
        return False

    def get_tag_dict(self):
        """Returns all tags as dict."""
        ret = {}
        tag_types = self.get_tag_types()
        for t in tag_types:
            ret[t] = self.get_tag_list(tag_type=t)
        return ret

    def get_tag_list(self, tag_type: str = "e"):
        """Returns all tags of given type as list."""
        ret = []
        for _tag_type, _tag in self.tags:
            if _tag_type == tag_type:
                ret.append(_tag)
        return ret

    def get_tag_types(self):
        """Returns list of all included tag types."""
        ret = []
        for tag_type, _ in self.tags:
            if tag_type not in ret:
                ret.append(tag_type)
        return ret

    def get_tag_count(self, tag_type: str = "e"):
        """Returns all tags of given type as list."""
        count = 0
        for _tag_type, _tag in self.tags:
            if _tag_type == tag_type:
                count += 1
        return count

    def verify(self) -> bool:
        pub_key = PublicKey.from_hex(self.public_key)
        return pub_key.verify(bytes.fromhex(self.signature), bytes.fromhex(self.id))

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "pubkey": self.public_key,
            "created_at": self.created_at,
            "kind": self.kind,
            "tags": self.tags,
            "content": self.content,
            "sig": self.signature,
        }

    @classmethod
    def from_dict(cls, msg: dict) -> "Event":
        return Event(
            content=msg["content"],
            public_key=msg["pubkey"],
            created_at=msg["created_at"],
            kind=msg["kind"],
            tags=msg["tags"],
            signature=msg["sig"],
        )

    def to_message(self) -> str:
        return json.dumps([ClientMessageType.EVENT, self.to_dict()])

    def __repr__(self):
        note_id = self.bech32()
        return f"Event({note_id[:10]}...{note_id[-10:]})"

    def __str__(self):
        return self.to_message()


@dataclass
class EncryptedDirectMessage(Event):
    recipient_pubkey: str = None
    cleartext_content: str = None
    reference_event_id: str = None

    def __post_init__(self):
        if self.content:
            self.cleartext_content = self.content
            self.content = None

        if self.recipient_pubkey is None:
            raise Exception("Must specify a recipient_pubkey.")

        self.kind = EventKind.ENCRYPTED_DIRECT_MESSAGE
        super().__post_init__()

        # Must specify the DM recipient's pubkey in a 'p' tag
        self.add_pubkey_ref(self.recipient_pubkey)

        # Optionally specify a reference event (DM) this is a reply to
        if self.reference_event_id:
            self.add_event_ref(self.reference_event_id)

    @property
    def id(self) -> str:
        if self.content is None:
            raise Exception(
                "EncryptedDirectMessage `id` is undefined \
                until its message is encrypted and stored in the `content` field"
            )
        return super().id



import json
import unittest
from unittest import mock
from unittest.mock import ANY, MagicMock, patch

from click.testing import CliRunner

from nostr.commands.message import cli
from nostr.event import Event
from nostr.message_pool import EventMessage


class TestCLIMessage(unittest.TestCase):
    @patch('nostr.commands.message.RelayManager', autospec=True)
    def test_publish(self, mock_relay_manager):
        # GIVEN
        nsec = "nsec1lrjqzalcev9ard0274pu8ynwx0xzzexh56sfn0c97rumh8f2tfcqd3lf8h"
        message = "Hello nostr world!"
        runner = CliRunner()

        mock_manager = MagicMock()
        mock_relay_manager.return_value = mock_manager
        mock_manager.__enter__.return_value = mock_manager
        mock_manager.add_relay.return_value = None

        # WHEN
        result = runner.invoke(
            cli,
            ['publish', '-s', nsec, '-m', message, '--sleep', 0],
            catch_exceptions=False,
        )

        # THEN
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(json.loads(result.output), {"Message": message})
        mock_manager.add_relay.assert_has_calls([mock.call(ANY), mock.call(ANY)])
        mock_manager.publish_message.assert_called()

    @patch('nostr.commands.message.RelayManager', autospec=True)
    def test_send(self, mock_relay_manager):
        # GIVEN
        nsec = "nsec1jqaxtwk4tymddju9dmn58tdxgwgl5ck23gg847fla9cyuslxklrq86fcjd"
        npub = "npub1mg2nzunrsk9df94zr3uudhzltnu6lzq2muax09xmhu5gxxrvnkqsvpjg3p"
        message = "Hello nostr world!"
        runner = CliRunner()

        mock_manager = MagicMock()
        mock_relay_manager.return_value = mock_manager
        mock_manager.__enter__.return_value = mock_manager
        mock_manager.add_relay.return_value = None

        # WHEN
        result = runner.invoke(
            cli,
            ['send', '-s', nsec, '-m', message, '-p', npub, '--sleep', 0],
        )

        # THEN
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(json.loads(result.output), {"Message": message})
        mock_manager.add_relay.assert_has_calls([mock.call(ANY), mock.call(ANY)])
        mock_manager.publish_event.assert_called()

    @patch('nostr.commands.message.RelayManager', autospec=True)
    def test_receive(self, mock_relay_manager):
        # GIVEN
        npub = "npub1mg2nzunrsk9df94zr3uudhzltnu6lzq2muax09xmhu5gxxrvnkqsvpjg3p"
        runner = CliRunner()

        mock_manager = MagicMock()
        mock_relay_manager.return_value = mock_manager
        mock_manager.__enter__.return_value = mock_manager
        mock_manager.add_relay.return_value = None

        mock_manager.message_pool.has_events.side_effect = [True, False]
        mock_manager.message_pool.get_event.return_value = EventMessage(
            Event(content="any"), subscription_id=123, url="wss://any.relay"
        )

        mock_manager.message_pool.has_notices.return_value = False

        # WHEN
        result = runner.invoke(
            cli,
            ['receive', '-p', npub, '-s', 0],
        )

        # THEN
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(
            json.loads(result.output),
            {"Public key(s)": [npub], "Events": ANY, "Notices": ANY},
        )
        mock_manager.add_relay.assert_has_calls([mock.call(ANY), mock.call(ANY)])
        mock_manager.publish_message.assert_called()
        mock_manager.message_pool.has_events.assert_called()
        mock_manager.message_pool.get_event.assert_called()
        mock_manager.message_pool.has_notices.assert_called()
        mock_manager.message_pool.get_notice.assert_not_called()



import json
from collections import UserList
from typing import List

from nostr.event import Event, EventKind


class Filter:
    """NIP-01 filtering.

    Explicitly supports "#e" and "#p" tag filters via `event_refs` and
    `pubkey_refs`.

    Arbitrary NIP-12 single-letter tag filters are also supported via
    `add_arbitrary_tag`.
    If a particular single-letter tag gains prominence, explicit support
    should be
    added. For example:
    # arbitrary tag
    filter.add_arbitrary_tag('t', [hashtags])

    # promoted to explicit support
    Filter(hashtag_refs=[hashtags])
    """

    def __init__(
        self,
        event_ids: List[str] = None,
        kinds: List[EventKind] = None,
        authors: List[str] = None,
        since: int = None,
        until: int = None,
        event_refs: List[
            str
        ] = None,  # the "#e" attr; list of event ids referenced in an "e" tag
        pubkey_refs: List[
            str
        ] = None,  # The "#p" attr; list of pubkeys referenced in a "p" tag
        limit: int = None,
    ) -> None:
        self.event_ids = event_ids
        self.kinds = kinds
        self.authors = authors
        self.since = since
        self.until = until
        self.event_refs = event_refs
        self.pubkey_refs = pubkey_refs
        self.limit = limit

        self.tags = {}
        if self.event_refs:
            self.add_arbitrary_tag("e", self.event_refs)
        if self.pubkey_refs:
            self.add_arbitrary_tag("p", self.pubkey_refs)

    def add_arbitrary_tag(self, tag: str, values: list):
        """Filter on any arbitrary tag with explicit handling for NIP-01 and NIP-12
        single-letter tags."""
        # NIP-01 'e' and 'p' tags and
        # any NIP-12 single-letter tags must be prefixed with "#"
        tag_key = tag if len(tag) > 1 else f"#{tag}"
        self.tags[tag_key] = values

    @classmethod
    def from_json(cls, filters):
        if "ids" in filters:
            ret = cls(filters["ids"])
        else:
            ret = cls("")
        if "authors" in filters:
            ret.authors = filters["authors"]
        if "kinds" in filters:
            ret.kinds = filters["kinds"]
        if "#e" in filters:
            ret.event_refs = filters["#e"]
        if "#p" in filters:
            ret.pubkey_refs = filters["#p"]
        if "since" in filters:
            ret.since = filters["since"]
        if "until" in filters:
            ret.until = filters["until"]
        if "limit" in filters:
            ret.limit = filters["limit"]
        cls.tags = {}
        if cls.event_refs:
            cls.add_arbitrary_tag("e", cls.event_refs)
        if cls.pubkey_refs:
            cls.add_arbitrary_tag("p", cls.pubkey_refs)
        return ret

    def matches(self, event: Event) -> bool:
        if self.event_ids and event.id not in self.event_ids:
            return False
        if self.kinds and event.kind not in self.kinds:
            return False
        if self.authors and event.public_key not in self.authors:
            return False
        if self.since and event.created_at < self.since:
            return False
        if self.until and event.created_at > self.until:
            return False
        if (self.event_refs or self.pubkey_refs) and len(event.tags) == 0:
            return False

        if self.tags:
            e_tag_identifiers = {e_tag[0] for e_tag in event.tags}
            for f_tag, f_tag_values in self.tags.items():
                # Omit any NIP-01 or NIP-12 "#" chars on single-letter tags
                f_tag = f_tag.replace("#", "")

                if f_tag not in e_tag_identifiers:
                    # Event is missing a tag type that we're looking for
                    return False

                # Multiple values within f_tag_values are treated as OR search;
                # an Event needs to match only one.
                # Note: an Event could have multiple entries of the same tag type
                # (e.g. a reply to multiple people) so we have to check all of them.
                match_found = False
                for e_tag in event.tags:
                    if e_tag[0] == f_tag and e_tag[1] in f_tag_values:
                        match_found = True
                        break
                if not match_found:
                    return False

        return True

    def to_json_object(self) -> dict:
        res = {}
        if self.event_ids:
            res["ids"] = self.event_ids
        if self.kinds:
            res["kinds"] = self.kinds
        if self.authors:
            res["authors"] = self.authors
        if self.since:
            res["since"] = self.since
        if self.until:
            res["until"] = self.until
        if self.limit:
            res["limit"] = self.limit
        if self.tags:
            res.update(self.tags)

        return res

    def __repr__(self):
        return f"Filters({self.to_json_object()})"

    def __str__(self):
        return json.dumps(self.to_json_object())


class Filters(UserList):
    def __init__(self, initlist: "list[Filter]" = None) -> None:
        super().__init__(initlist)
        self.data: "list[Filter]"

    def match(self, event: Event):
        for filter in self.data:
            if filter.matches(event):
                return True
        return False

    def to_json_array(self) -> list:
        """Convert the data of the object to a json array."""
        return [filter.to_json_object() for filter in self.data]

    def __repr__(self):
        return f"FilterList({self.to_json_array()})"

    def __str__(self):
        return json.dumps(self.to_json_array())



import json
from dataclasses import dataclass
from pathlib import Path

import hcl2


@dataclass
class Config:
    filename: str = "config.hcl"

    @classmethod
    def locate(cls) -> Path:
        """
        Locate config path

        # Search paths:
        # 1. ${CURRENT_PATH}/config.hcl
        # 2. ${HOME}/.nostr/config.hcl
        """
        paths = [
            Path.cwd().joinpath(cls.filename),
            Path.home().joinpath('.nostr', cls.filename),
        ]

        for path in paths:
            if path.exists():
                return path

        return None

    @classmethod
    def load(cls, filename: str = None) -> dict:
        """
        Load the config file
        """
        filename = filename or cls.locate()
        if filename:
            with open(filename) as file:
                return hcl2.load(file)
        return {}

    @staticmethod
    def dump(content: dict = None) -> str:
        """
        Dump the config content
        """
        if content:
            return json.dumps(content, indent=2)



import unittest
from unittest.mock import mock_open, patch

from nostr.commands.config import Config


class TestConfig(unittest.TestCase):
    @patch('nostr.commands.config.Path')
    def test_locate_with_existing_file_in_cwd(self, mock_path):
        # GIVEN
        expected_path = mock_path('/cwd/path/to/config.hcl')
        # Mock the cwd method to return a known path
        mock_path.cwd.return_value.joinpath.return_value = expected_path
        mock_path.cwd.return_value.joinpath.return_value.exists.return_value = True

        mock_path.home.return_value.joinpath.return_value.exists.return_value = False

        # WHEN
        actual_path = Config.locate()
        # THEN
        self.assertEqual(expected_path, actual_path)
        mock_path.cwd.return_value.joinpath.assert_called_once_with('config.hcl')

    @patch('nostr.commands.config.Path')
    def test_locate_with_existing_file_in_home_directory(self, mock_path):
        # GIVEN
        expected_path = mock_path('/home/path/to/config.hcl')
        # Mock the home method to return a known path
        mock_path.home.return_value.joinpath.return_value = expected_path
        mock_path.home.return_value.joinpath.return_value.exists.return_value = True

        mock_path.cwd.return_value.joinpath.return_value.exists.return_value = False
        # WHEN
        actual_path = Config.locate()
        # THEN
        self.assertEqual(expected_path, actual_path)
        mock_path.home.return_value.joinpath.assert_called_once_with(
            '.nostr', 'config.hcl'
        )

    @patch('nostr.commands.config.Path')
    def test_locate_file_not_found(self, mock_path):
        # GIVEN
        mock_path.cwd.return_value.joinpath.return_value.exists.return_value = False
        mock_path.home.return_value.joinpath.return_value.exists.return_value = False
        # WHEN
        actual = Config.locate()
        # THEN
        self.assertEqual(None, actual)

    @patch('nostr.commands.config.Config.locate', return_value="test.hcl")
    @patch('builtins.open', new_callable=mock_open, read_data='key = "value"')
    def test_load_valid_file(self, mock_file, mock_locate):
        config = Config.load()
        mock_file.assert_called_once_with("test.hcl")
        self.assertEqual(config, {'key': 'value'})
        mock_locate.assert_called_once_with()

    @patch('nostr.commands.config.Config.locate')
    @patch('builtins.open', new_callable=mock_open, read_data='key = "value"')
    def test_load_valid_file_with_param(self, mock_file, mock_locate):
        config = Config.load("test.hcl")
        mock_file.assert_called_once_with("test.hcl")
        self.assertEqual(config, {'key': 'value'})
        mock_locate.assert_not_called()

    @patch('nostr.commands.config.Config.locate', return_value=None)
    @patch('builtins.open', side_effect=FileNotFoundError)
    def test_load_file_not_found(self, mock_file, mock_locate):
        config = Config.load()
        self.assertEqual(config, {})
        mock_locate.assert_called_once_with()

    def test_dump_with_content(self):
        content = {"key": "value"}
        result = Config.dump(content)
        expected_result = "{\n  \"key\": \"value\"\n}"
        self.assertEqual(result, expected_result)

    def test_dump_without_content(self):
        result = Config.dump()
        self.assertIsNone(result)
